freeze;

declare attributes FldFin: Witt_ring;

Z2 := Integers(2);
Z4 := Integers(4);
C2 := AbelianGroup([2]);

//////////////////////////////////////////////////
// IsSumOfSquares
// (just for completeness)
intrinsic IsSumOfSquares( a::FldFinElt ) -> BoolElt
{ " }
    return true;
end intrinsic;

intrinsic IsSOS( a::FldFinElt ) -> BoolElt
{ " }
    return IsSumOfSquares(a);
end intrinsic;

//////////////////////////////////////////////////
// LengthOfSOS
// (just for completeness)
intrinsic LengthOfSOS( a::FldFinElt ) -> RngIntElt
{
Given a nonzero element a, this function determines
minimal number of summands needed to express a as a sum of squares. 
It returns 0 if a is not a sum of squares
}
    return IsSquare(a) select 1 else 2;
end intrinsic;

intrinsic LengthOfSumOfSquares( a::FldFinElt ) -> RngIntElt
{ " } //"
    return LengthOfSOS(a);
end intrinsic;


//////////////////////////////////////////////////
// Level

intrinsic Level( K::FldFin ) -> RngIntElt
{    
This function comutes the level s(K) of a number field K,
i.e. a lowest number s such that -1 is a sum of s squares
it returns 0 if K is formally real.
}
    s := case< #K mod 4 | 3 : 2, default: 1 >;
    assert2 s eq LengthOfSOS(K!-1);
    return s;
end intrinsic;

//////////////////////////////////////////////////
// IsFormallyReal
// just for completeness
intrinsic IsFormallyReal( K::FldRat ) -> BoolElt
{
Returns true iff K is formally real (i.e. -1 is not a sum of squares)
}
    return false;
end intrinsic;

intrinsic IsReal( K::FldRat ) -> BoolElt
{ " } //"
    return false;
end intrinsic;


//////////////////////////////////////////////////
// PythagorasNumber & PythagorasElement

intrinsic PythagorasNumber( K::FldFin ) -> RngIntElt
{
This function computes the Pythagoras number P(K) of a field K,
i.e. the lowest number p such that every sum of squares in K
is a sum of p squares.
}
    // Use [Szymiczek (1997), Exm. 18.2.1] and
    // [Lam (2005), Prop. II.3.4]
    return IsOdd(Characteristic(K)) select 2 else 1;
end intrinsic;

intrinsic PythagorasElement( K::FldFin ) -> FldNumElt
{
Returns an element a of K whose length equals the Pythagoras 
number P(K) of K.
}
    return IsOdd(Characteristic(K)) select Nonsquare(K) else One(K);
end intrinsic;

//////////////////////////////////////////////////
// quadratic forms

function is_isotropic_FF( q )
    // Let q be a diagonal quadratic form with coefficients in
    // a finite field of odd characteristic. This function checks
    // if q is isotropic
    
    // the base field
    F := q`base_ring;
    assert IsFinite(F);
    assert Characteristic(F) ne 2;
    
    d := Dimension(q);
    if d le 1 then
        return false;
    elif d eq 2 then
        t := IsSquare(Discriminant(q));
        return t;
    end if;
    return true;
end function;

function is_hyperbolic_FF( q )
    // Let q be a diagonal quadratic form with coefficients in
    // a finite field of odd characteristic. This function checks
    // if q is hyperbolic
    
    // the base field
    F := q`base_ring;
    assert IsFinite(F);
    assert Characteristic(F) ne 2;
    
    d := Dimension(q);
    if IsOdd(d) then
        return false;
    end if;
    if IsSquare(Discriminant(q)) then
        return true;
    end if;
    return false;
end function;


//////////////////////////////////////////////////
// Witt ring
intrinsic WittRing( F::FldFin ) -> RngWittQF
{
Constructs the Witt ring of similarity classes of quadratic forms over F.
}
    if not assigned F`Witt_ring then
        W := New(RngWittQF);
        W`base_ring := F;
        W`is_explicit := true;
        if #F mod 4 eq 1 then
            // Use [Szymiczek (1997), Exc. 13.3.5]
            W`explicit_model := GroupAlgebra(Z2, C2);
        elif #F mod 4 eq 3 then
            // Use [Szymiczek (1997), Exc. 13.3.5]
            W`explicit_model, _ := Z4;
        elif Characteristic(F) eq 2 then
            // Use [Szymiczek (1997), Prop. 13.2.3]
            W`explicit_model, _ := Z2;
        else 
            require false: "The base field is not supported!";
        end if;
        F`Witt_ring := W;
    end if;
    return F`Witt_ring;
end intrinsic;

//////////////////////////////////////////////////
// IsTorsion
// just for completeness
function is_torsion_FF( _q )
// Check if the element _q of a Witt ring is torsion.
    return true;
end function;
