freeze;

declare attributes DiagQuadFrm: primes;

// TODO: cache primes that matter for q

declare attributes FldNum: Lvl, PythNum, PythNumElt, WittEqInv, Hght;

WittEqInv_fmt := recformat<
    d,  // absolute degree`
    r,  // number of real embeddings
    s,  // level
    k,  // number of dyadic places
    L   // list of pairs (degree, level) for all dyadic places
>;

ZZ := Integers();
QQ := Rationals();

//////////////////////////////////////////////////
// IsLocalSquare

intrinsic IsLocalSquare( a::FldNumElt, P::PlcNumElt ) -> BoolElt
{ 
Given an element a in K, decide if a is a square in the completion K_P
}
    return IsLocalSquare(a, Ideal(P));
end intrinsic;


//////////////////////////////////////////////////
// IsSumOfSquares

intrinsic IsSumOfSquares( a::FldNumElt ) -> BoolElt
{
Returns true iff a is a sum of squares
}
    //return IsTotallyPositive(a);
    return forall{ s : s in RealSigns(a) | s eq 1 };
end intrinsic;

intrinsic IsSOS( a::FldNumElt ) -> BoolElt
{ " } //"
    return IsSumOfSquares(a);
end intrinsic;


//////////////////////////////////////////////////
// LengthOfSOS

function length_nondyadic( a, P )
// Given a nonzero element a of a number field K and a nondyadic place P, 
// this fuction determines the minimal number of summands needed to express 
// a as a sum of squares in a completion K_P.  
    
    // the residue field
    Kp := ResidueClassField(P);
    // and valuation
    v := Valuation(a,P);
    // two cases to consider
    if IsOdd(v) then
        // possible values are 2 and 3
        q := DiagonalQuadraticForm(Kp, [Kp!1, Kp!1]);
        return IsIsotropic(q) select 2 else 3;
    else
        // possible values are 1 and 2
        return IsLocalSquare(a, P) select 1 else 2;
    end if;
end function;

function length_dyadic( a, P )
// Given a nonzero element a of a number field K and a dyadic place P, 
// this fuction determines the minimal number of summands needed to express 
// a as a sum of squares in a completion K_P.  

    K := Parent(a);
    // if it is a square, the answer is trivial
    if IsLocalSquare(a, P) then
        return 1;
    end if;
    // now possible values are 2, 3, 4
    // check for 2
    if HilbertSymbol(K!-1, a, Ideal(P)) eq 1 then
        return 2;
    end if;
    // the two possibilities left are 3 and 4
    // check 3
    if HilbertSymbol(K!-1, K!-1, Ideal(P)) eq 1 or not IsLocalSquare(-a, P) then
        return 3;
    end if;
    // now the only choice is 4
    return 4;
end function;

intrinsic LengthOfSOS( a::FldNumElt ) -> RngIntElt
{
Given a nonzero element a of a number field K, this function determines
minimal number of summands needed to express a as a sum of squares. 
It returns 0 if a is not a sum of squares
}
    require not IsZero(a): "The argument must be nonzero!";
    // start with trivial cases
    if not IsSOS(a) then
        return 0;
    end if;
    if IsSquare(a) then
        return 1;
    end if;
    // the possible values now are 2, 3, 4
    K := Parent(a);
    OK := MaximalOrder(K);
    // find all the primes of K that matter for a
    even_primes := [ P[1] : P in Decomposition(K, 2) ];
    //odd_primes :=  [ P[1] : P in Factorization(a*OK) ];
    odd_primes, _ := [ P : P in Support(Divisor(a)) | P notin even_primes ];
    vprint DQF_verbosity : "Primes to be tested:", even_primes, odd_primes;
    
    vprint DQF_verbosity : "Lengths at dydadic primes:", [ length_dyadic(a, P) : P in even_primes ];
    m := Max([ length_dyadic(a, P) : P in even_primes ]);
    vprint DQF_verbosity : "m = ", m;
    vprint DQF_verbosity : "Lengths at non-dydadic primes:", [ length_dyadic(a, P) : P in odd_primes ];
    n := IsEmpty(odd_primes) select -1 else Max([ length_nondyadic(a, P) : P in odd_primes ]);
    vprint DQF_verbosity : "n =", n;
    l := Max([2, m, n]);
    return l;
end intrinsic;

intrinsic LengthOfSumOfSquares( a::FldNumElt ) -> RngIntElt
{ " } //"
    return LengthOfSOS(a);
end intrinsic;


//////////////////////////////////////////////////
// LegendreSymbol

intrinsic LegendreSymbol( a::FldNumElt, P::PlcNumElt ) -> RngIntElt
{ 
Given an element a and a place P of a number field, 
this function computes the Legendre symbol a over P. 
}
    v := Valuation(a, P);
    if IsOdd(v) then
        return 0;
    end if;
    if v ne 0 then
        p := UniformizingElement(P);
        a := a*p^(-v);
    end if;
    if IsSquare(Evaluate(a,P)) then
        return 1;
    end if;
    return -1;
end intrinsic;

intrinsic LegendreSymbol( a::FldNumElt, P::RngOrdIdl ) -> RngIntElt
{ " }
    require IsPrime(P): "P must be a prime ideal!";
    return LegendreSymbol(a, Place(P));
end intrinsic;


//////////////////////////////////////////////////
// HasseMinkowskiInvariant
// WittInvariant

intrinsic HasseMinkowskiInvariant( q::DiagQuadFrm, P::RngOrdIdl ) -> RngIntElt
{
The Hasse invariant of a diagonal quadratic form q = < a1, ..., ad> \
at a prime P is a product of Hilbert symbols (c.f. [Lam (2005), Def. V.3.17]) \
prod_(i<j) (ai, aj)_p.
This function computes the Hasse symbol of q at p.
}
    require IsPrime(P): "P must be a prime ideal!";
    require q`base_ring eq NumberField(Order(P)): "Base field of P does not match base field of q!";

    d := Dimension(q);
    L := [ HilbertSymbol(q`coeffs[i],q`coeffs[j], P) : j in [(i+1)..d] , i in [1..d-1] ];
    return &* L;
end intrinsic;

intrinsic HasseMinkowskiInvariant( q::DiagQuadFrm, P::PlcNumElt ) -> RngIntElt
{ " } //"
    return HasseMinkowskiInvariant(q, Ideal(P));
end intrinsic;

intrinsic WittInvariant( q::DiagQuadFrm, P::RngOrdIdl ) -> RngIntElt
{ " } //"
    return HasseMinkowskiInvariant(q, P);
end intrinsic;
    
intrinsic WittInvariant( q::DiagQuadFrm, P::PlcNumElt ) -> RngIntElt
{ " } //"
    return HasseMinkowskiInvariant(q, Ideal(P));
end intrinsic;


//////////////////////////////////////////////////
// IsIsotropic

function is_locally_isotropic_nondyadic( q, P )
// Given a quadratic form q over a number field or QQ and a prime ideal P
// this function checks if the localization of q at p is isotropic.

    // unary forms cannot be isotropic
    if Dimension(q) le 1 then
        return false;
    end if;
    // quintic and higher dimensional forms are always isotropic
    if Dimension(q) ge 5 then
        return true;
    end if;

    // the base field and its maximal order
    K := q`base_ring;
    OK := MaximalOrder(K);
  
    // the residue field
    Kp, kappa := ResidueClassField(P);
    // uniformizer
    if ISA(Category(OK), RngInt) then
        p := Generator(P);
    else 
        p := UniformizingElement(P);
    end if;

    // split the form into q = q0 + p*q1
    q0 := [ a*p^(-Valuation(a,P)) : a in q`coeffs | IsEven(Valuation(a,P)) ];
    q1 := [ a*p^(-Valuation(a,P)) : a in q`coeffs | IsOdd(Valuation(a,P)) ];
    q0 := [ a*Denominator(a)^2 : a in q0 ];
    q1 := [ a*Denominator(a)^2 : a in q1 ];
    
    // take the reduction mod gp
    q0bar := [ kappa(a) : a in q0 ];
    q1bar := [ kappa(a) : a in q1 ];
    // verify if q0, q1 are isotropic
    A := IsIsotropic( DiagonalQuadraticForm(Kp, q0bar) );
    B := IsIsotropic( DiagonalQuadraticForm(Kp, q1bar) );
    return A or B;
end function;

function is_locally_isotropic_dyadic( q, P )
// Given a quadratic form q over a number field and a prime ideal P
// this function checks if the localization of q at p is isotropic.

    // the base field and its maximal order
    K := q`base_ring;

    // the dimension
    d := Dimension(q);
    // the determinant
    detq := &* q`coeffs;
    //Kd, phi := Completion(Order(P), P);
    
    if d le 1 then
        return false;
    elif d ge 5 then
        return true;
    elif d eq 2 then
        return IsLocalSquare(-detq, P);
    elif d eq 3 then
        // use [Lam, Prop. V.3.22]
        s := HilbertSymbol(K!-1, -detq, P);
        sq := WittInvariant(q, P);
        return (s eq sq);
    elif IsLocalSquare(detq, P) then
        // if we get here, then d == 4        
        // for quartic form with determinant 1, use [ Lam, Proposition V.3.23]
        s := HilbertSymbol(K!-1, K!-1, P);
        sq := WittInvariant(q, P);
        return (s eq sq);
    end if;
    // form is unique with det a square
    return true;
end function;

intrinsic IsIsotropic( q::DiagQuadFrm, P::RngOrdIdl ) -> BoolElt
{
Checks if a quadratic form q over a number field is locally isotropic 
at P.
}
    require IsPrime(P): "P must be a prime ideal!";
    require q`base_ring eq NumberField(Order(P)): "Base field of P does not match base field of q!";
    K := q`base_ring;
    if (K!2) notin P then
        return is_locally_isotropic_nondyadic(q,P);
    else
        return is_locally_isotropic_dyadic(q,P);
    end if;
end intrinsic;

intrinsic IsIsotropic( q::DiagQuadFrm, P::PlcNumElt ) -> BoolElt
{ " } //"
    return IsIsotropic(q, Ideal(P));
end intrinsic;

function signatures_builtin( q )
    // Given a quadratic form q over a number field K, this function returns
    // a sequence of signatures of q at all real embedings of K.
    // This implementation uses a built-in function RealSigns.
    
    K := q`base_ring;
    assert ISA(Category(K), FldNum);
    
    r, _ := Signature(K);
    S := [ RealSigns(c) : c in q`coeffs ];
    return [ &+ [ s[j] : s in S ] : j in [1..r] ];
end function;

function is_isotropic_NF( q )
    // Given a quadratic form q with coefficients in a number field K,
    // this function tests if q is isotropic.

    // the base field and its maximal order
    K := q`base_ring;
    OK := MaximalOrder(K);

    // the dimension
    d := Dimension(q);
    if d le 1 then
        return false;
    elif d eq 2 then
        return IsSquare(Discriminant(q));
    end if;

    // compute the signatures of q in all real embedding of K
    S := signatures_builtin( q );
    vprint DQF_verbosity : "Signatures: ", S;
    // check if q is isotropic at every real place
    if not &and [ Abs(s) lt d : s in S ] then
        return false;
    end if;
    vprint DQF_verbosity : "Signatures passed";
          
    // find all the primes of K that matter for q
    if not assigned q`primes then
        even_primes := [ Ideal(P[1]) : P in Decomposition(K, 2) ];
        odd_primes := &join [ {a[1] : a in Factorization(c*OK)} : c in q`coeffs ];
        odd_primes := [ P : P in odd_primes ];
        q`primes := even_primes cat odd_primes;
    end if;
    vprint DQF_verbosity : "Primes to be tested:", q`primes;
    // check that q is locally isotropic over all of them
    return forall{ P : P in q`primes | IsIsotropic(q, P) };
end function;


//////////////////////////////////////////////////
// IsHyperbolic

intrinsic IsHyperbolic( q::DiagQuadFrm, P::PlcNumElt ) -> BoolElt
{
Given a quadratic form q over a number field K and a prime P of K, 
this function returns true iff q is hyperbolic in the completion
of K at P.
}
    d := Dimension(q);
    // hyperbolic forms are always even-dimensional
    if IsOdd(d) then
        return false;
    end if;
    // and their discriminants are squares
    if not IsLocalSquare(Discriminant(q), P) then
        return false;
    end if;
    m := ZZ ! (d/2);
    K := q`base_ring;
    t := HilbertSymbol(K!-1,K!-1,Ideal(P))^( ZZ ! (m^2-m)/2 );
    return HasseMinkowskiInvariant(q, P) eq t;
end intrinsic;

intrinsic IsHyperbolic( q::DiagQuadFrm, P::RngOrdIdl ) -> BoolElt
{ " } //"
    require IsPrime(P): "P must be a prime ideal!";
    require q`base_ring eq NumberField(Order(P)): "Base field of P does not match base field of q!";
    return IsHyperbolic(q, Place(P));
end intrinsic;

function is_hyperbolic_NF( q )
    // Given a quadratic form q over a number field K, 
    // this function returns true if q is hyperbolic and false otherwise.
    
    // the base field and its maximal order
    K := q`base_ring;
    OK := MaximalOrder(K);

    d := Dimension(q);
    // hyperbolic forms are always even-dimensional
    if IsOdd(d) then
        return false;
    end if;
    // and their discriminants are squares
    if not IsSquare(Discriminant(q)) then
        return false;
    end if;

    // compute the signatures of q in all real embedding of K
    S := signatures_builtin( q );
    // if it contains a nonzero, then q is non-hyperbolic at this embedding
    if exists{s : s in S | not IsZero(s) } then
        return false;
    end if;

    // find all the primes of K that matter for q
    if not assigned q`primes then
        even_primes := [ Ideal(P[1]) : P in Decomposition(K, 2) ];
        odd_primes := &join [ {a[1] : a in Factorization(c*OK)} : c in q`coeffs ];
        odd_primes := [ P : P in odd_primes ];
        q`primes := even_primes cat odd_primes;
    end if;
    vprint DQF_verbosity : "Primes to be tested:", q`primes;
    
    // check that q is locally hyperbolic over all of them
    return forall{ P : P in q`primes | IsHyperbolic(q, P) };
end function;


//////////////////////////////////////////////////
// AnisotropicDimension

intrinsic AnisotropicDimension( q::DiagQuadFrm, P::RngOrdIdl ) -> RngIntElt
{
Every non-degenerate quadratic form q decomposes (uniquely up to an isometry) 
as q ~ q_1 + ... + c*(1, -1), where q_1 is anisotropic. The form q_1 is called an anisotropic part of q.

Given a quadratic form q and a prime P, this function returns a dimension 
of an anisotropic part of localilization of q at P.
}
    require IsPrime(P): "P must be a prime ideal!";
    require q`base_ring eq NumberField(Order(P)): "Base field of P does not match base field of q!";
    
    d := Dimension(q);
    if IsEven(d) then
        // dimension of hyperbolic form is 0
        if IsHyperbolic(q, P) then
            return 0;
        elif IsLocalSquare(Discriminant(q), P) then
            return 4;
        else 
            return 2;
        end if;
    end if;
    // when q is odd-dimensional, extend it to the next even dimension
    c := QQ ! (-1)^(ZZ ! (d*(d+1)/2));
    q2 := DiagonalQuadraticForm( q`base_ring, q`coeffs cat [ c * &* q`coeffs ] );
    if IsHyperbolic( q2, P ) then
        return 1;
    else
        return 3;
    end if;
end intrinsic;

intrinsic AnisotropicDimension( q::DiagQuadFrm, P::PlcNumElt ) -> RngIntElt
{ " } // "
    return AnisotropicDimension(q, Ideal(P));
end intrinsic;

function anisotropic_dimension_NF( q )
// Every non-degenerate quadratic form q decomposes (uniquely up to an isometry) 
// as q ~ q_1 + ... + c*(1, -1), where q_1 is anisotropic. 
// The form q_1 is called an anisotropic part of q.
//
// Given a quadratic form q over a number field K, 
// this function returns a dimension of an anisotropic part of q.

    // the base field and its maximal order
    K := q`base_ring;
    OK := MaximalOrder(K);

    // compute the signatures of q in all real embedding of K
    S := signatures_builtin( q );
    vprint DQF_verbosity : "Signatures: ", S;
    if not IsEmpty(S) then
        N := Max([Abs(s) : s in S]);
    end if;

    // for n > 2, it must be a real place, where the the anisotropic
    // part has a maximal dimension
    if N  ge 3 then
        return N;
    end if;
    vprint DQF_verbosity : "Signatures passed";
          
    // find all the primes of K that matter for q
    if not assigned q`primes then
        even_primes := [ Ideal(P[1]) : P in Decomposition(K, 2) ];
        odd_primes := &join [ {a[1] : a in Factorization(c*OK)} : c in q`coeffs ];
        odd_primes := [ P : P in odd_primes ];
        q`primes := even_primes cat odd_primes;
    end if;
    vprint DQF_verbosity : "Primes to be tested:", q`primes;
    // compute maximal anisotropic dimension over all primes
    M :=  Max([ AnisotropicDimension(q, P) : P in q`primes ]);
    return Max(M, N);    
end function;


//////////////////////////////////////////////////
// IsFormallyReal

intrinsic IsFormallyReal( K::FldNum ) -> BoolElt
{
Returns true iff K is formally real (i.e. -1 is not a sum of squares)
}
    r,c := Signature(K);
    return r ne 0;
end intrinsic;

intrinsic IsReal( K::FldNum ) -> BoolElt
{ " } //"
    return IsFormallyReal(K);
end intrinsic;


//////////////////////////////////////////////////
// Level

intrinsic Level( K::FldNum ) -> RngIntElt
{    
This function comutes the level s(K) of a number field K,
i.e. a lowest number s such that -1 is a sum of s squares
it returns 0 if K is formally real.
}
    if assigned K`Lvl then
        return K`Lvl;
    end if;

    // level of real field is 0 (or infinity)
    if IsFormallyReal(K) then
        K`Lvl := 0;
        return 0;
    end if;
    // is level equal 1
    if IsSquare(K!-1) then
        K`Lvl := 1;
        return 1;
    end if;
    // now level is either 2 or 4 depending on dyadic primes
    D := Decomposition(K, 2);
    for d in D do
        f := InertiaDegree(d[1]);
        e := RamificationIndex(d[1]);
        if IsOdd(f) and IsOdd(e) then
            K`Lvl := 4;
            return 4;
        end if;
    end for;
    K`Lvl := 2;
    return 2;
end intrinsic;


//////////////////////////////////////////////////
// PythagorasNumber & PythagorasElement

intrinsic PythagorasNumber( K::FldNum ) -> RngIntElt
{
This function computes the Pythagoras number P(K) of a number field K,
i.e. the lowest number p such that every sum of squares in K
is a sum of p squares.
}
    if assigned K`PythNum then
        return K`PythNum;
    end if;

    // if s(K)=1 then P(K)=2
    if IsSquare(K!-1) then
        K`PythNum := 2;
        return K`PythNum;
    end if;
    // now level is either 3 or 4 depending on dyadic primes
    D := Decomposition(K, 2);
    for d in D do
        f := InertiaDegree(d[1]);
        e := RamificationIndex(d[1]);
        if IsOdd(f) and IsOdd(e) then
            K`PythNum := 4;
            return K`PythNum;
        end if;
    end for;
    K`PythNum := 3;
    return K`PythNum;
end intrinsic;

intrinsic PythagorasElement( K::FldNum ) -> FldNumElt
{
Returns an element a of K whose length equals the Pythagoras 
number P(K) of K.
}
    if assigned K`PythNumElt then
        return K`PythNumElt;
    end if;
    
    PK := PythagorasNumber(K);
    sK := Level(K);
    if PK eq 2 then
        // K is nonreal by [Lam 2005, Exm. XI.5.9] 
        // and its level must be 1 by [Koprowski, Czogała 2018, Prop. 4]
        K`PythNumElt := -1;
    elif PK eq 3 then
        // Level is 0 or 2
        p := 2;
        while not assigned K`PythNumElt do
            D := Decomposition(K, p);
            for P in D do
                if not IsSquare( Evaluate(K!-1, P[1]) ) then
                    // this is ad hoc correction, more work needed
                    a := UniformizingElement( P[1] );
                    if IsSOS(a) then 
                        K`PythNumElt := UniformizingElement( P[1] );
                        break;
                    end if;
                end if;
            end for;
            p := NextPrime(p);
        end while;
    elif PK eq 4 then
        // level is either 0 or 4
        K`PythNumElt := K!7;
    else
        error if false, "Something went wrong!";
    end if;
    assert2 LengthOfSOS(K`PythNumElt) eq PythagorasNumber(K);
    return K`PythNumElt;
end intrinsic;

//////////////////////////////////////////////////
// Height

intrinsic Height( K::FldNum ) -> RngIntElt
{
This function computes the height h(K) of a number field K,
i.e. exp( Wt(K) ), where Wt(K) is denotes the torsion subgroup
of the Witt ring of K.
}
    if assigned K`Hght then
        return K`Hght;
    end if;
    
    if not IsFormallyReal(K) then
        // if K is non-real, then h(K) = 2*s(K)
        // see [Lam, p. 395]
        K`Hght := 2*Level(K);
    else
        // otherwise it is 4 by
        // [Lam, Thm 5.6] and posible values of P(K)
        K`Hght := 4;
    end if;
    return K`Hght;
end intrinsic;


//////////////////////////////////////////////////
// IsTorsion

function is_torsion_NF( _q )
// Check if the element _q of a Witt ring is torsion.

    q := _q`form;
    // compute the signatures of q in all real embedding of K
    S := signatures_builtin( q );
    // use Szymiczek (1997), Cor. 17.3.1 and Cor. 19.2.3
    return forall{ IsZero(s) : s in S };
end function;

//////////////////////////////////////////////////
// AreWittEquivalent

function local_dyadic_level( d )
    // For a dyadic place d of a number field K, 
    // this function computes the level s(Kd) of 
    // the completion Kd of K at d
    
    K := NumberField(d);
    // if s(K)=1 then s(Kd) = 1
    if IsSquare(K!-1) then
        return 1;
    end if;
    
    f := InertiaDegree(d);
    e := RamificationIndex(d);
    if IsOdd(e) and IsOdd(f) then
        return 4;
    end if;
    if IsOdd(e) and IsEven(f) then
        return 2;
    end if;
    I := Ideal(d);
    R := Order(I);
    Kd, phi := Completion(R, I);
    if IsSquare(phi(-1)) then
        return 1;
    end if;
    return 2;
end function;

function Witt_eq_invariant( K )
    // Given a number field K, this function computes its
    // complete set of Witt equivalence invariants
    
    if not assigned K`WittEqInv then
        r, _ := Signature(K);
        D := Decomposition(K, 2);
        L := [ <LocalDegree(d), local_dyadic_level(d)> where d := d_[1] : d_ in D ]; 
        Sort(~L);
        K`WittEqInv := rec< WittEqInv_fmt | 
            d := AbsoluteDegree(K),
            r := r,
            s := Level(K),
            k := #D,
            L := L
        >;
    end if;
    return <K`WittEqInv`d, K`WittEqInv`r, K`WittEqInv`s, K`WittEqInv`k, K`WittEqInv`L>;
end function;

intrinsic AreWittEquivalent( K::FldNum, L::FldNum ) -> BoolElt
{
Given two fields K and L, this function return true if their Witt rings are isomorphic and false otherwise.
}    
    WK := Witt_eq_invariant(K);
    WL := Witt_eq_invariant(L);
    return WK eq WL;
end intrinsic;

intrinsic AreWittEquivalent( K::FldNum, L::FldRat ) -> BoolElt
{ " } //"
    return AbsoluteDegree(K) eq 1;
end intrinsic;

intrinsic AreWittEquivalent( K::FldNum, L::FldFun[FldFin] ) -> BoolElt
{ " } //"
    // Use Szymiczek (1991), Theorem 1.3
    return false;
end intrinsic;

intrinsic AreWittEquivalent( K::FldNum, L::FldFunRat ) -> BoolElt
{ " } //"
    require ISA(Category(ConstantField(L)), FldFin) : "Not implemented for these types of fields!";
    // Use Szymiczek (1991), Theorem 1.3
    return false;
end intrinsic;







